from functools import lru_cache
from typing import Union


def group_len(type_keyword: bytes) -> int:
    """
    The length of element groups in unformatted arrays.
    """
    if type_keyword[0:1] == b"C":
        return 105
    else:
        return 1000


# The number of bytes for each
# res type except C0nn which varies
static_item_sizes = {
    b"INTE": 4,
    b"REAL": 4,
    b"LOGI": 4,
    b"DOUB": 8,
    b"CHAR": 8,
    b"MESS": 0,
    b"X231": 0,
}


@lru_cache
def item_size(type_keyword: bytes) -> Union[int, None]:
    """
    :returns: The number of bytes for each element in an
        res array of the given res type.
    """
    if type_keyword[0:2] == b"C0":
        return int(type_keyword[2:4].decode("ascii"))
    return static_item_sizes.get(type_keyword)


@lru_cache
def bytes_in_array(array_length: int, item_type: bytes) -> Union[int, None]:
    """
    :param array_length: Number of items in the array
    :param item_type: Type of items in the array
    :returns: Number of bytes used to store an array of
        given type and length
    """
    g_len = group_len(item_type)
    full_groups = array_length // g_len

    size = item_size(item_type)
    if size is None:
        return None

    if array_length % g_len:
        return (full_groups + 1) * 8 + array_length * size
    else:
        return full_groups * 8 + array_length * size



import io

import numpy as np

import resfo._formatted.write as fwrite
from resfo.types import MESS


def test_write_str_list():
    buf = io.StringIO()
    fwrite.write_str_list(buf, ["a", "bb"])

    assert buf.getvalue() == "\n 'a ' 'bb'"


def test_write_str():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("ZGRP    ", ["12345678"] * 8)])

    assert (
        buf.getvalue()
        == """ 'ZGRP    '           8 'CHAR'
 '12345678' '12345678' '12345678' '12345678' '12345678' '12345678' '12345678'
 '12345678'\n"""
    )


def test_write_int():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("FILEHEAD", np.zeros(shape=(7,), dtype=np.int32))])

    assert (
        buf.getvalue()
        == """ 'FILEHEAD'           7 'INTE'
           0           0           0           0           0           0
           0\n"""
    )


def test_write_logi():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("LOGIHEAD", np.zeros(shape=(26,), dtype=np.bool_))])

    assert (
        buf.getvalue()
        == """ 'LOGIHEAD'          26 'LOGI'
  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F  F
  F\n"""
    )


def test_write_real():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("REALHEAD", np.zeros(shape=(4,), dtype=np.float32))])

    assert (
        buf.getvalue()
        == """ 'REALHEAD'           4 'REAL'
   0.00000000E+00   0.00000000E+00   0.00000000E+00   0.00000000E+00\n"""
    )


def test_write_doub():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("DOUBHEAD", np.zeros(shape=(4,), dtype=np.float64))])

    assert (
        buf.getvalue()
        == """ 'DOUBHEAD'           4 'DOUB'
   0.00000000000000D+00   0.00000000000000D+00   0.00000000000000D+00
   0.00000000000000D+00\n"""
    )


def test_write_mess():
    buf = io.StringIO()
    fwrite.formatted_write(buf, [("MESSHEAD", MESS)])
    assert buf.getvalue() == """ 'MESSHEAD'           0 'MESS'\n"""



import io

import numpy as np
import pytest

import resfo._unformatted.write as uwrite
from resfo.types import MESS


def test_write_array_header():
    buf = io.BytesIO()
    uwrite.write_array_header(buf, "KEYWORD1", b"INTE", 3)

    assert (
        buf.getvalue()
        == b"\x00\x00\x00\x10KEYWORD1\x00\x00\x00\x03INTE\x00\x00\x00\x10"
    )


def test_write_array_header_with_x231():
    buf = io.BytesIO()
    uwrite.write_array_header(buf, "KEYWORD1", b"INTE", 5000000000)

    marker = (16).to_bytes(4, byteorder="big", signed=True)
    assert buf.getvalue() == (
        marker
        + b"KEYWORD1"
        + (-2).to_bytes(4, byteorder="big", signed=True)
        + b"X231"
        + marker
        + marker
        + b"KEYWORD1"
        + (705032704).to_bytes(4, byteorder="big", signed=True)
        + b"INTE"
        + marker
    )


def test_write_array_header_short_keyword():
    buf = io.BytesIO()

    with pytest.raises(ValueError, match="must have length"):
        uwrite.write_array_header(buf, "KEY", b"INTE", 3)


def test_write_array_header_not_a_type():
    buf = io.BytesIO()

    with pytest.raises(uwrite.ResfoWriteError, match="a valid res type"):
        uwrite.write_array_header(buf, "KEYWORD1", b"NOTY", 3)


@pytest.mark.filterwarnings("ignore:casting")
@pytest.mark.parametrize(
    "array, expected_array",
    [
        (np.array([1], dtype=np.int32), np.array([1], dtype=np.int32)),
        (np.array([1.0], dtype=np.float32), np.array([1.0], dtype=np.float32)),
        (np.array([1.0], dtype=np.float64), np.array([1.0], dtype=np.float64)),
        (np.array([True, False], dtype=np.bool_), np.array([-1, 0], dtype=np.int32)),
        (np.array([10, 20], dtype=np.int8), np.array([10, 20], dtype=np.int32)),
        (np.array([1.0], dtype=np.longdouble), np.array([1.0], dtype=np.float64)),
    ],
)
def test_cast_array_to_res(array, expected_array):
    assert np.array_equal(uwrite.cast_array_to_res(array), expected_array)


def test_cast_array_to_res_bad_type():
    array = np.array([b"ABC"], dtype="|S")
    with pytest.raises(ValueError, match="Cannot cast"):
        uwrite.cast_array_to_res(array)


@pytest.mark.filterwarnings("ignore:casting")
def test_write_np_array():
    buf = io.BytesIO()
    uwrite.write_np_array(
        buf, np.ones(shape=(1300,), dtype=np.dtype(np.int32).newbyteorder(">"))
    )

    assert buf.getvalue() == (
        (
            (4000).to_bytes(4, byteorder="big", signed=True)
            + b"\x00\x00\x00\x01" * 1000
            + (4000).to_bytes(4, byteorder="big", signed=True)
        )
        + (
            (1200).to_bytes(4, byteorder="big", signed=True)
            + b"\x00\x00\x00\x01" * 300
            + (1200).to_bytes(4, byteorder="big", signed=True)
        )
    )


def test_write_str_list_too_long():
    buf = io.BytesIO()
    str_list = ["a" * 200]

    with pytest.raises(ValueError, match="strings of length"):
        uwrite.write_str_list(buf, str_list, b"CHAR")


def test_write_str_list_non_ascii():
    buf = io.BytesIO()
    str_list = ["\u2167"]

    with pytest.raises(ValueError, match="non-ascii"):
        uwrite.write_str_list(buf, str_list, b"CHAR")


def test_write_str_list_char():
    buf = io.BytesIO()
    str_list = ["a" * 8, "b" * 4]

    uwrite.write_str_list(buf, str_list, b"CHAR")

    marker = (16).to_bytes(4, byteorder="big", signed=True)

    assert buf.getvalue() == marker + b"a" * 8 + b"b" * 4 + b" " * 4 + marker


def test_write_mess():
    buf = io.BytesIO()

    uwrite.unformatted_write(buf, [("MESSHEAD", MESS)])

    marker = (16).to_bytes(4, byteorder="big", signed=True)

    assert (
        buf.getvalue() == marker + b"MESSHEAD" + b"\x00\x00\x00\x00" + b"MESS" + marker
    )



import warnings
from typing import IO, Any, Iterable, List, Tuple, Union

import numpy as np
import numpy.typing as npt

import resfo.types as res_types
from resfo._unformatted.common import group_len, item_size
from resfo.errors import ResfoWriteError
from resfo.types import MessType


def write_array_header(
    stream: IO[bytes], kw_str: str, type_str: bytes, size: int
) -> None:
    if len(kw_str) != 8:
        raise ResfoWriteError("keywords must have length exactly size 8")
    if not res_types.is_valid_type(type_str):
        raise ResfoWriteError(f"Not a valid res type: {str(type_str)}")

    if size > 2**31:
        write_array_header(stream, kw_str, b"X231", -(size // (2**31)))
        size %= 2**31

    stream.write((16).to_bytes(4, byteorder="big", signed=True))
    stream.write(kw_str.encode("ascii"))
    stream.write(size.to_bytes(4, byteorder="big", signed=True))
    stream.write(type_str)
    stream.write((16).to_bytes(4, byteorder="big", signed=True))


def cast_array_to_res(arr: npt.NDArray[Any]) -> npt.NDArray[Any]:
    if arr.dtype.type in [np.int32, np.float32, np.float64]:
        return arr.astype(arr.dtype.newbyteorder(">"))
    if np.issubdtype(arr.dtype, np.bool_):
        return -arr.astype(">i4")
    elif np.issubdtype(arr.dtype, np.integer):
        result_dtype = ">i4"
    elif np.issubdtype(arr.dtype, np.floating):
        result_dtype = ">f8"
    else:
        raise ValueError(f"Cannot cast {arr.dtype} to a res type")

    warnings.warn(f"casting array dtype {arr.dtype} to {result_dtype}", stacklevel=1)
    return arr.astype(result_dtype)


def write_np_array(stream: IO[bytes], arr: npt.ArrayLike) -> None:
    arr = np.asarray(arr)
    res_type = res_types.from_np_dtype(arr)
    if res_type == b"C008":
        raise ValueError()
    g_len = group_len(res_type)
    to_write = len(arr)
    type_len = item_size(res_type)
    if type_len is None:
        raise ValueError(f"Unknown type {str(res_type)}")
    have_written = 0
    arr = cast_array_to_res(arr)
    while to_write > 0:
        write_now = min(g_len, to_write)
        to_write -= write_now
        bytes_to_write = type_len * write_now
        stream.write(bytes_to_write.to_bytes(4, byteorder="big", signed=True))
        stream.write(arr[have_written : have_written + write_now].tobytes())
        have_written += write_now
        stream.write(bytes_to_write.to_bytes(4, byteorder="big", signed=True))


def write_str_list(stream: IO[bytes], str_list: List[str], res_type: bytes) -> None:
    str_size = item_size(res_type)
    if str_size is None:
        raise ValueError(f"Unknown type {str(res_type)}")
    if len(str_list) == 0:
        return
    max_len = max(len(s) for s in str_list)
    if max_len > 99:
        raise ResfoWriteError("Res files does not support strings of length > 99")
    if max_len > str_size:
        raise ResfoWriteError(
            f"Inconsistent type size, have {str_size} type but longest string is {max_len}"
        )
    str_list = [s.ljust(str_size) for s in str_list]
    try:
        byte_str_list = [
            s.encode("ascii") if isinstance(s, str) else s for s in str_list
        ]
    except UnicodeEncodeError as e:
        raise ResfoWriteError(
            "Cannot write non-ascii strings to unformatted res files"
        ) from e

    g_len = group_len(res_type)
    to_write = len(byte_str_list)
    type_len = item_size(res_type)
    if type_len is None:
        raise ValueError(f"Unknown type {type_len}")
    have_written = 0
    while to_write > 0:
        write_now = min(g_len, to_write)
        to_write -= write_now
        bytes_to_write = type_len * write_now
        stream.write(bytes_to_write.to_bytes(4, byteorder="big", signed=True))
        for i in range(have_written, have_written + write_now):
            stream.write(byte_str_list[i])
        have_written += write_now
        stream.write(bytes_to_write.to_bytes(4, byteorder="big", signed=True))


def write_array_like(
    stream: IO[bytes], keyword: str, array_like: Union[npt.ArrayLike, MessType]
) -> None:
    array = np.asarray(array_like)
    try:
        res_type = res_types.from_np_dtype(array)
    except ValueError as e:
        raise ResfoWriteError(f"{e}") from e
    if res_type == b"MESS":
        write_array_header(stream, keyword, res_type, 0)
        array = np.array([])
    else:
        write_array_header(stream, keyword, res_type, len(array))
    if np.issubdtype(array.dtype, np.str_) or array.dtype.char == "S":
        write_str_list(stream, array.tolist(), res_type)
    else:
        write_np_array(stream, array)


def unformatted_write(
    stream: IO[bytes],
    keyworded_arrays: Iterable[Tuple[str, Union[npt.ArrayLike, MessType]]],
) -> None:
    for keyword, array in keyworded_arrays:
        write_array_like(stream, keyword, array)



import io
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike

import resfo.types as res_types
from resfo._unformatted.common import bytes_in_array, group_len, item_size
from resfo._unformatted.write import write_array_like
from resfo.array_entry import ResArray
from resfo.errors import ResfoParsingError


class UnformattedResArray(ResArray):
    """
    An array entry in a unformatted res file.
    """

    def read_array(self) -> res_types.ArrayValue:
        """
        Read the array from the unformatted res file.
        :returns: numpy array of values.
        """
        if self._data_start is None:
            self._read()
        if self._type == b"MESS":
            return res_types.MESS
        assert self._data_start is not None
        assert self._type is not None
        assert self._length is not None
        self.stream.seek(self._data_start)
        g_len = group_len(self._type)
        np_type = res_types.to_np_type(self._type)
        array = np.zeros(shape=(self._length,), dtype=np_type)
        to_read = self._length
        type_len = item_size(self._type)
        if type_len is None:
            raise ResfoParsingError(f"Unknown type {str(type_len)}")
        have_read = 0
        while to_read > 0:
            read_now = min(g_len, to_read)
            to_read -= read_now
            bytes_to_read = type_len * read_now
            self._read_record_marker(bytes_to_read)
            try:
                group_array = np.frombuffer(
                    self.stream.read(bytes_to_read), dtype=np_type, count=read_now
                )
            except ValueError as error:
                raise ResfoParsingError(f"Could not read array: {error}") from error
            array[have_read : have_read + read_now] = group_array
            have_read += read_now
            self._read_record_marker(bytes_to_read)

        if self._type == b"LOGI":
            array = array.astype(np.bool_)

        return array

    def update(
        self, *, keyword: Optional[str] = None, array: Optional[ArrayLike] = None
    ) -> None:
        if keyword is None:
            keyword = self.read_keyword()

        array_: res_types.ArrayValue = res_types.MESS
        if not isinstance(array, res_types.MessType):
            array_ = np.asarray(array) if array is not None else self.read_array()
        if (
            not isinstance(array_, res_types.MessType)
            and self.read_length() != array_.size
        ):
            raise ValueError("Cannot update array with different size")

        if self.read_type() != res_types.from_np_dtype(array_):
            raise ValueError("Cannot update array with different type")

        self.stream.seek(self.start)
        write_array_like(self.stream, keyword, array_)
        self._keyword = keyword
        self._array = array_

    def _read_record_marker(self, expected_value: int) -> None:
        """
        Read 4 bytes and check that it matches the
        expected fortan header/footer
        """
        value = int.from_bytes(self.stream.read(4), byteorder="big", signed=True)
        if value != expected_value:
            raise ResfoParsingError(
                f"Unexpected size of record {value} ({str(value.to_bytes(4, byteorder='big', signed=True))})"
            )

    def _read_keyword(self) -> None:
        """
        With stream.peek() at the start of the keyword, reads
        it into self._keyword
        """
        self._keyword = self.stream.read(8).decode("ascii")
        if not self._keyword or len(self._keyword) < 8:
            raise ResfoParsingError("Reached end-of-file while reading keyword")

    def _read_type(self) -> None:
        """
        With stream.peek() at the start of the type string, reads
        it into self._type
        """
        self._type = self.stream.read(4)
        if not self._type or len(self._type) < 4:
            raise ResfoParsingError("Reached end-of-file while reading type")

    def _read_length(self) -> int:
        """
        With stream.peek() at the start of the length, reads it and returns it.
        """
        length_bytes = self.stream.read(4)
        if not length_bytes or len(length_bytes) < 4:
            raise ResfoParsingError("Reached end-of-file while reading length")

        return int.from_bytes(length_bytes, byteorder="big", signed=True)

    def _read(self) -> None:
        """
        See resfo.array_entry.ResArray._read()
        """
        self.stream.seek(self.start)
        start_marker = self.stream.read(4)
        if not start_marker:
            self._is_eof = True
            return
        start_marker_value = int.from_bytes(start_marker, byteorder="big", signed=True)
        if start_marker_value != 16:
            raise ResfoParsingError(
                f"Unexpected size of record {start_marker_value} ({start_marker})"
            )
        self._read_keyword()
        self._length = self._read_length()
        self._read_type()
        self._read_record_marker(16)
        if self._type == b"X231":
            self._read_record_marker(16)
            self._length *= -(2**31)
            previous_keyword = self._keyword
            self._read_keyword()
            if previous_keyword != self._keyword:
                raise ResfoParsingError(
                    "x231 type record was not followed "
                    f"by record with same keyword, found {self._keyword}"
                    f" expected {previous_keyword}"
                )
            self._length += self._read_length()
            self._read_type()
            self._read_record_marker(16)

        type_len = item_size(self._type)
        if self._length == 0:
            type_len = 0
        if type_len is None:
            raise ResfoParsingError(f"Unexpected item type {str(self._type)}")
        if type_len <= 0 and self._length != 0:
            raise ResfoParsingError(
                f"Found keyword of type {str(self._type)} which"
                f"has item length {type_len} which requires 0 number of"
                f"elements, but found {self._length} amount of elements."
            )
        bytes_to_skip = bytes_in_array(self._length, self._type)
        if bytes_to_skip is None:
            raise ResfoParsingError(f"Unknown type {str(self._type)}")

        self._data_start = self.stream.tell()
        self.stream.seek(bytes_to_skip, io.SEEK_CUR)



import io

import numpy as np
import pytest

from resfo._unformatted.read import UnformattedResArray
from resfo._unformatted.write import unformatted_write


@pytest.mark.filterwarnings("ignore:casting")
@pytest.mark.filterwarnings("ignore:downcasting")
@pytest.mark.parametrize(
    "data",
    [
        [("KEYWORD1", [1, 2, 3])],
        [("KEYWORD1", [""])],
        [
            ("KEYWORD1", [1, 2, 3]),
            ("22222222", np.array([1.1, 2.1, 3.1])),
        ],
        [
            ("KEYWORD1", np.array([1, 2, 3])),
            ("KEYWORD2", [1.0, 2.0, 3.0]),
        ],
        [
            ("KEYWORS1", np.array(["HELLO WORLD"], dtype="|S")),
            ("KEYWORS2", np.array(["HELLO WORLD"], dtype="|S11")),
            ("KEYWORU1", np.array(["HELLO WORLD"], dtype="|U")),
            ("KEYWORU2", np.array(["HELLO WORLD"], dtype="|U11")),
        ],
    ],
)
def test_unformatted_write(data):
    buf = io.BytesIO()
    unformatted_write(buf, data)
    buf.seek(0)
    out_records = list(UnformattedResArray.parse(buf))
    out_keywords = [r.read_keyword() for r in out_records]
    out_arrays = [r.read_array() for r in out_records]
    out_lengths = [r.read_length() for r in out_records]

    assert out_lengths == [len(arr) for arr in out_arrays]

    for (kw, arr), okw, oarr in zip(data, out_keywords, out_arrays, strict=True):
        assert kw == okw
        if isinstance(arr[0], str):
            for el, elo in zip(arr, oarr, strict=True):
                elo = elo.decode("ascii")
                assert elo.startswith(el)
                assert all(c.isspace() for c in elo[len(el) :])
        else:
            assert np.array_equal(arr, oarr)



import hypothesis.strategies as st
import numpy as np
from hypothesis.extra.numpy import arrays

import resfo

formats = st.sampled_from(resfo.Format)
in_formats = st.one_of(formats, st.just(None))
keywords = st.text(
    min_size=8, max_size=8, alphabet=st.characters(min_codepoint=40, max_codepoint=126)
)

unicode_arrays = st.builds(np.array, st.lists(keywords))
str_arrays = st.builds(lambda arr: arr.astype("|S"), unicode_arrays)
int_arrays = arrays(dtype=np.int32, shape=(10,))
float_arrays = arrays(
    dtype=np.float64,
    elements=st.floats(width=64, min_value=-1e100, max_value=1e100),
    shape=(10,),
)

res_arrays = st.one_of(
    float_arrays, int_arrays, str_arrays, unicode_arrays, st.just(resfo.MESS)
)
resfo_datas = st.lists(st.tuples(keywords, res_arrays))



class ResfoParsingError(ValueError):
    """
    Indicates an error occurred during reading of an res file.
    """

    pass


class ResfoWriteError(ValueError):
    """
    Indicates an error occurred during writing of an res file.
    """

    pass



from typing import IO, Any, Iterable, List, Tuple, Union

import numpy as np
import numpy.typing as npt

import resfo.types as res_types
from resfo.errors import ResfoWriteError
from resfo.types import MessType


def write_str_list(stream: IO[str], str_list: List[str]) -> None:
    """
    Writes the given list of strings to the res file.
    The strings will have the length of the longest string,
    left-padded with space and enclosed by single-quotes.
    """
    if len(str_list) == 0:
        return
    max_len = max(len(s) for s in str_list)
    str_list = [s.decode("ascii") if not isinstance(s, str) else s for s in str_list]

    for i, string in enumerate(str_list):
        if i % 7 == 0:
            stream.write("\n")
        stream.write(f" '{string.ljust(max_len)}'")


def write_np_array(stream: IO[str], array: Iterable[Any], res_type: bytes) -> None:
    """
    Writes the given numpy array to the stream as
    the given res_type.
    :param stream: stream to write to.
    :param array: Numpy array.
    :param res_type: The res type to use for the array, see
        resfo.types.
    """
    if res_type == b"LOGI":
        for i, ele in enumerate(array):
            if i % 25 == 0:
                stream.write("\n")
            if ele:
                stream.write("  T")
            else:
                stream.write("  F")
    elif res_type == b"REAL":
        for i, ele in enumerate(array):
            if i % 4 == 0:
                stream.write("\n")
            stream.write(" {:>16.8E}".format(ele))
    elif res_type == b"DOUB":
        for i, ele in enumerate(array):
            if i % 3 == 0:
                stream.write("\n")
            stream.write(" {:>22.14E}".format(ele).replace("E", "D"))
    elif res_type == b"INTE":
        for i, ele in enumerate(array):
            if i % 6 == 0:
                stream.write("\n")
            stream.write(" {:>11d}".format(ele))


def write_entry(
    stream: IO[str], keyword: str, array_like: Union[npt.ArrayLike, MessType]
) -> None:
    """
    Write the given array/keyword entry to the
    stream in the formatted res file format.
    :param stream: text-mode stream to write to.
    :param keyword: 8-character string to use for keyword.
    :param array_like: Array of values to write.
    """
    if "'" in keyword:
        raise ResfoWriteError('keywords in formatted files cannot contain "\'"')
    array = np.asarray(array_like)
    try:
        res_type = res_types.from_np_dtype(array)
    except ValueError as e:
        raise ResfoWriteError(f"{e}") from e
    if res_type == b"MESS":
        stream.write(
            f" '{keyword.ljust(8)}' {' {:>10d}'.format(0)} '{res_type.decode('ascii').ljust(4)}'"
        )
    else:
        stream.write(
            f" '{keyword.ljust(8)}' {' {:>10d}'.format(len(array))} '{res_type.decode('ascii').ljust(4)}'"
        )

    if np.issubdtype(array.dtype, np.str_) or array.dtype.char == "S":
        write_str_list(stream, array.tolist())
    else:
        write_np_array(stream, array, res_type)


def formatted_write(
    stream: IO[str],
    keyworded_arrays: Iterable[Tuple[str, Union[npt.ArrayLike, MessType]]],
) -> None:
    """
    Writes the list of data entries to the stream as
    formatted res.
    :param stream: File handle in text mode.
    :param keyworded_arrays: Either iterable or list of (kw, array)
        tuples for entries to be written to the file. Also takes dictionary
        of keyword: array pairs.
    """
    iterator = keyworded_arrays
    for keyword, array in iterator:
        write_entry(stream, keyword, array)
        stream.write("\n")



from os import PathLike
from typing import IO, Any, Iterator, List, Optional, Tuple, Union

from resfo._formatted.read import FormattedArray
from resfo._unformatted.read import UnformattedResArray
from resfo.array_entry import ResArray
from resfo.format import Format, check_correct_mode, get_stream, guess_format

from .types import ArrayValue


def read(
    filelike: Union[str, PathLike[str], IO[Any]], fileformat: Optional[Format] = None
) -> List[Tuple[str, ArrayValue]]:
    """
    Read the contents of a res file and return a list of
    tuples (keyword, array). Takes the same parameters as
    lazy_read, but differs in return type
    """
    return [
        (arr.read_keyword(), arr.read_array())
        for arr in lazy_read(filelike, fileformat)
    ]


def lazy_read(
    filelike: Union[str, PathLike[str], IO[Any]], fileformat: Optional[Format] = None
) -> Iterator[ResArray]:
    """
    Reads the contents of an res file and generates the entries
    of that file. Each entry has a entry.read_keyword() and
    entry.read_array() method which will return the corresponding
    data, but only upon request. This requires the user to
    pay attention to when values are read as it should happen
    before the file is closed.

    When lazy_read is given a path or filename, the file will be closed once
    the generator has ran out of elements.

    For greater control, one can pass an opened file so that close
    can be called at the correct time.

    :param filelike: Either filename, pathlib.Path or stream
        to write file to. For fileformat=Format.UNFORMATTED the
        stream must be in binary mode and for fileformat=Format.FORMATTED
        in text mode.
    :param fileformat: Either resfo.Format.FORMATTED for ascii
        format, resfo.Format.UNFORMATTED for binary formatted files
        or None for guess.

    :raises resfo.ResfoParsingError: If the file is not a valid
        res file.

    .. note::
        If given a file to be open (as opposed to a stream), the errors
        (various `IOError` s) associated with the default behavior of the
        built-in `open()` function may be raised.

        When given a stream, the exceptions associated with the stream will
        pass through.
    """
    if fileformat is None:
        fileformat = guess_format(filelike)
    stream, didopen = get_stream(filelike, fileformat)

    check_correct_mode(stream, fileformat)

    try:
        if fileformat == Format.FORMATTED:
            yield from FormattedArray.parse(stream)
        else:
            yield from UnformattedResArray.parse(stream)
    finally:
        if didopen:
            stream.close()



import numpy as np
import pytest

import resfo


@pytest.mark.parametrize(
    "contents, expected_type",
    [
        ("'SPECGRID' 1 'INTE' 1 ", np.dtype(np.int32).newbyteorder(">")),
        ("'SPECGRID' 1 'REAL' 1.0 ", np.dtype(np.float32).newbyteorder(">")),
        ("'SPECGRID' 1 'DOUB' 1.0 ", np.dtype(np.float64).newbyteorder(">")),
        ("'SPECGRID' 1 'CHAR' 'SPECGRID' ", np.dtype("<U8")),
        ("'SPECGRID' 1 'C010' 'SPECGRID99' ", np.dtype("<U10")),
        ("'SPECGRID' 1 'LOGI' T ", bool),
    ],
)
def test_formatted_read_type_resolution(tmp_path, contents, expected_type):
    with (tmp_path / "test.txt").open("w") as f:
        f.write(contents)

    with (tmp_path / "test.txt").open("r") as f:
        ((_specgrid, arr),) = resfo.read(f)

    assert arr.dtype == expected_type


@pytest.mark.parametrize(
    "data, expected_type",
    [
        (["KEYWORD1"], "'CHAR'"),
        (["KEYWORD10"], "'C009'"),
        ([1], "'INTE'"),
        ([1.0], "'DOUB'"),
        (np.array([1.0], dtype=np.float32), "'REAL'"),
        ([True], "'LOGI'"),
    ],
)
def test_formatted_write_type_resolution(tmp_path, data, expected_type):
    file = tmp_path / "test.txt"

    with file.open("w") as f:
        resfo.write(f, [("SPECGRID", data)], resfo.Format.FORMATTED)

    assert expected_type in file.read_text()


def keyword_start(res_type):
    return b"\x00\x00\x00\x10SPECGRID\x00\x00\x00\x01" + res_type + b"\x00\x00\x00\x10"


@pytest.mark.parametrize(
    "contents, expected_type",
    [
        (
            keyword_start(b"INTE")
            + b"\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04",
            np.dtype(np.int32).newbyteorder(">"),
        ),
        (
            keyword_start(b"REAL")
            + b"\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04",
            np.dtype(np.float32).newbyteorder(">"),
        ),
        (
            keyword_start(b"DOUB")
            + b"\x00\x00\x00\x08"
            + b"\x00" * 8
            + b"\x00\x00\x00\x08",
            np.dtype(np.float64).newbyteorder(">"),
        ),
        (
            keyword_start(b"CHAR")
            + b"\x00\x00\x00\x08"
            + b"SPECGRID"
            + b"\x00\x00\x00\x08",
            np.dtype("|S8"),
        ),
        (
            keyword_start(b"C010")
            + b"\x00\x00\x00\x0a"
            + b"SPECGRID10"
            + b"\x00\x00\x00\x0a",
            np.dtype("|S10"),
        ),
        (
            keyword_start(b"LOGI")
            + b"\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04",
            bool,
        ),
    ],
)
def test_unformatted_read_type_resolution(tmp_path, contents, expected_type):
    with (tmp_path / "test.txt").open("wb") as f:
        f.write(contents)

    with (tmp_path / "test.txt").open("rb") as f:
        ((_specgrid, arr),) = resfo.read(f)

    assert arr.dtype == expected_type


@pytest.mark.parametrize(
    "data, expected_type",
    [
        (["KEYWORD1"], b"CHAR"),
        (["KEYWORD10"], b"C009"),
        ([1], b"INTE"),
        ([1.0], b"DOUB"),
        (np.array([1.0], dtype=np.float32), b"REAL"),
        ([True], b"LOGI"),
    ],
)
def test_unformatted_write_type_resolution(tmp_path, data, expected_type):
    file = tmp_path / "test.txt"

    with file.open("wb") as f:
        resfo.write(f, [("SPECGRID", data)], resfo.Format.UNFORMATTED)

    assert expected_type in file.read_bytes()



import numpy as np

import resfo.types


def test_resfo_to_np_type():
    assert resfo.types.to_np_type(b"C032") == np.dtype("|S32")
    assert resfo.types.to_np_type(b"CHAR") == np.dtype("|S8")
    assert resfo.types.to_np_type(b"INTE") == np.dtype(">i4")
    assert resfo.types.to_np_type(b"REAL") == np.dtype(">f4")
    assert resfo.types.to_np_type(b"LOGI") == np.dtype(">i4")
    assert resfo.types.to_np_type(b"DOUB") == np.dtype(">f8")
    assert resfo.types.to_np_type(b"MESS") is None
    assert resfo.types.to_np_type(b"x321") is None


def test_from_np_dtype():
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype("|S32"))) == b"C032"
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype("|S8"))) == b"CHAR"
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype(">i4"))) == b"INTE"
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype(">f4"))) == b"REAL"
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype(np.bool_))) == b"LOGI"
    assert resfo.types.from_np_dtype(np.array([], dtype=np.dtype(">f8"))) == b"DOUB"
    assert resfo.types.from_np_dtype(resfo.types.MESS) == b"MESS"
    assert (
        resfo.types.from_np_dtype(np.array(["HELLO WORLD"], dtype=np.dtype("|S")))
        == b"C011"
    )



import io

import numpy as np
import pytest

import resfo
import resfo._unformatted.read as uread


def test_unformatted_array_eof():
    buf = io.BytesIO(b"")
    arr = uread.UnformattedResArray(buf)
    assert arr.is_eof


@pytest.fixture
def simple_unformatted_buffer():
    return io.BytesIO(
        b"\x00\x00\x00\x10KEYWORD1\x00\x00\x00\x01INTE\x00\x00\x00\x10"
        b"\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04"
    )


def test_unformatted_array_simple_read(simple_unformatted_buffer):
    arr = uread.UnformattedResArray(simple_unformatted_buffer)
    assert not arr.is_eof
    assert arr.read_keyword() == "KEYWORD1"
    assert arr.read_type() == b"INTE"
    assert arr.read_length() == 1
    assert np.array_equal(arr.read_array(), np.zeros(shape=(1,), dtype=">i4"))


def test_unformatted_array_simple_type_loads(simple_unformatted_buffer):
    arr = uread.UnformattedResArray(simple_unformatted_buffer)
    assert arr.read_type() == b"INTE"


def test_unformatted_array_simple_length_loads(simple_unformatted_buffer):
    arr = uread.UnformattedResArray(simple_unformatted_buffer)
    assert arr.read_length() == 1


def test_unformatted_array_big_read():
    buf = io.BytesIO(
        b"\x00\x00\x00\x10KEYWORD1"
        + (2300).to_bytes(4, byteorder="big")
        + b"INTE\x00\x00\x00\x10"
        + (
            (4000).to_bytes(4, byteorder="big")
            + b"\x00\x00\x00\x01" * 1000
            + (4000).to_bytes(4, byteorder="big")
        )
        * 2
        + (
            (4 * 300).to_bytes(4, byteorder="big")
            + b"\x00\x00\x00\x01" * 300
            + (4 * 300).to_bytes(4, byteorder="big")
        )
    )
    arr = uread.UnformattedResArray(buf)
    assert not arr.is_eof
    assert arr.read_keyword() == "KEYWORD1"
    assert arr.read_type() == b"INTE"
    assert arr.read_length() == 2300
    assert np.array_equal(arr.read_array(), np.ones(shape=(2300,), dtype=">i4"))


def test_unformatted_array_bad_record_marker():
    buf = io.BytesIO(
        b"\x00\x00\x00\x10KEYWORD1"
        + (2300).to_bytes(4, byteorder="big")
        + b"INTE\x00\x00\x00\x10"
        + b"\x00" * (4 * 2300 + 4 * 6)
    )
    arr = uread.UnformattedResArray(buf)
    with pytest.raises(uread.ResfoParsingError, match="size of record"):
        arr.read_array()


def test_unformatted_array_read_record_marker():
    buf = io.BytesIO(b"\x00\x00\x00\x00")
    arr = uread.UnformattedResArray(buf)

    arr._read_record_marker(0)
    assert not buf.read(1)
    buf.seek(0)

    with pytest.raises(uread.ResfoParsingError, match="size of record"):
        arr._read_record_marker(1)


def test_unformatted_array_bad_initial_marker():
    buf = io.BytesIO(b"\x00\x00\x00\x01KEYWORD1")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(uread.ResfoParsingError, match="size of record"):
        arr.read_keyword()


def test_unformatted_array_bad_middle_marker():
    buf = io.BytesIO(b"\x00\x00\x00\x10KEYWORD1\x00\x00\x00\x00INTE\x00\x00\x00\x01")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(uread.ResfoParsingError, match="size of record"):
        arr.read_keyword()


def test_unformatted_array_eof_during_keyword():
    buf = io.BytesIO(b"\x00\x00\x00\x10A")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(
        uread.ResfoParsingError, match="end-of-file while reading keyword"
    ):
        arr.read_keyword()


def test_unformatted_array_eof_during_length():
    buf = io.BytesIO(b"\x00\x00\x00\x10AKEYWORD\x00")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(
        uread.ResfoParsingError, match="end-of-file while reading length"
    ):
        arr.read_keyword()


def test_unformatted_array_eof_during_type():
    buf = io.BytesIO(b"\x00\x00\x00\x10AKEYWORD\x00\x00\x00\x00T")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(uread.ResfoParsingError, match="end-of-file while reading type"):
        arr.read_keyword()


@pytest.mark.parametrize(
    "typestr, size, value, pyvalue",
    [
        (b"INTE", 4, b"\x00" * 4, 0),
        (b"REAL", 4, b"\x00" * 4, 0.0),
        (b"LOGI", 4, b"\x00" * 4, 0),
        (b"DOUB", 8, b"\x00" * 8, 0.0),
        (b"CHAR", 8, b"A8STRING", b"A8STRING"),
        (b"C011", 11, b"HELLO WORLD", b"HELLO WORLD"),
    ],
)
def test_unformatted_array_types(typestr, size, value, pyvalue):
    buf = io.BytesIO(
        b"\x00\x00\x00\x10AKEYWORD\x00\x00\x00\x01"
        + typestr
        + b"\x00\x00\x00\x10"
        + size.to_bytes(4, signed=True, byteorder="big")
        + value
        + size.to_bytes(4, signed=True, byteorder="big")
    )
    arr = uread.UnformattedResArray(buf)

    assert arr.read_keyword() == "AKEYWORD"
    assert np.array_equal(arr.read_array(), np.array([pyvalue]))


def test_unformatted_array_non_type():
    buf = io.BytesIO(b"\x00\x00\x00\x10AKEYWORD\x00\x00\x00\x01TYPE\x00\x00\x00\x10")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(uread.ResfoParsingError, match="Unexpected item type"):
        arr.read_keyword()


def test_unformatted_array_bad_end_marker():
    buf = io.BytesIO(b"\x00\x00\x00\x10AKEYWORD\x00\x00\x00\x01INTE\x00\x00\x10\x10")
    arr = uread.UnformattedResArray(buf)

    with pytest.raises(uread.ResfoParsingError, match="size of record"):
        arr.read_keyword()


def test_unformatted_array_zero_len():
    buf = io.BytesIO(b"\x00\x00\x00\x10AKEYWORD\x00\x00\x00\x00INTE\x00\x00\x00\x10")
    arr = uread.UnformattedResArray(buf)
    assert arr.read_keyword() == "AKEYWORD"
    assert len(arr.read_array()) == 0


def test_unformatted_simple_unformatted_parse():
    buf = io.BytesIO(
        (
            b"\x00\x00\x00\x10KEYWORD1\x00\x00\x00\x01INTE\x00\x00\x00\x10"
            b"\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04"
        )
        * 10
    )
    for arr in uread.UnformattedResArray.parse(buf):
        assert not arr.is_eof
        assert arr.read_keyword() == "KEYWORD1"
        assert arr.read_type() == b"INTE"
        assert arr.read_length() == 1
        assert np.array_equal(arr.read_array(), np.zeros(shape=(1,), dtype=">i4"))

    buf.seek(0)

    assert len(list(uread.UnformattedResArray.parse(buf))) == 10


def test_unformatted_x231_parse():
    """
    This test simply tests the math of the X231 reading, but
    relies on the lazy evaluation to avoid having an array
    with 5,000,000,000 elements follow the two records
    defining the array.
    """
    buf = io.BytesIO(
        (
            (16).to_bytes(4, byteorder="big", signed=True)
            + b"KEYWORD1"
            + (-2).to_bytes(4, byteorder="big", signed=True)
            + b"X231"
            + (16).to_bytes(4, byteorder="big", signed=True)
            + (16).to_bytes(4, byteorder="big", signed=True)
            + b"KEYWORD1"
            + (705032704).to_bytes(4, byteorder="big", signed=True)
            + b"INTE"
            + (16).to_bytes(4, byteorder="big", signed=True)
        )
    )
    for arr in uread.UnformattedResArray.parse(buf):
        assert not arr.is_eof
        assert arr.read_keyword() == "KEYWORD1"
        assert arr.read_type() == b"INTE"
        assert arr.read_length() == 5000000000

    buf.seek(0)

    assert len(list(uread.UnformattedResArray.parse(buf))) == 1


def test_unformatted_x231_wrong_kw():
    buf = io.BytesIO(
        (
            (16).to_bytes(4, byteorder="big", signed=True)
            + b"KEYWORD1"
            + (-1).to_bytes(4, byteorder="big", signed=True)
            + b"X231"
            + (16).to_bytes(4, byteorder="big", signed=True)
            + (16).to_bytes(4, byteorder="big", signed=True)
            + b"KEYWORD2"
            + (1).to_bytes(4, byteorder="big", signed=True)
            + b"INTE"
            + (16).to_bytes(4, byteorder="big", signed=True)
        )
    )
    with pytest.raises(uread.ResfoParsingError, match="x231 type record"):
        list(uread.UnformattedResArray.parse(buf))


def test_unformatted_parse_mess():
    buf = io.BytesIO(
        (
            (16).to_bytes(4, byteorder="big", signed=True)
            + b"KEYWORD1"
            + (0).to_bytes(4, byteorder="big", signed=True)
            + b"MESS"
            + (16).to_bytes(4, byteorder="big", signed=True)
        )
    )

    for arr in uread.UnformattedResArray.parse(buf):
        assert not arr.is_eof
        assert arr.read_keyword() == "KEYWORD1"
        assert arr.read_type() == b"MESS"
        assert arr.read_length() == 0

    buf.seek(0)

    assert len(list(uread.UnformattedResArray.parse(buf))) == 1


def test_truncated_file_raises_parse_error():
    buf = io.BytesIO()
    resfo.write(buf, [("KEYWORD1", np.ones(100))])
    buf.seek(0)
    buf.truncate(100)
    with pytest.raises(resfo.ResfoParsingError, match="Could not read array"):
        _ = resfo.read(buf)



from typing import IO, Optional

import numpy as np
from numpy.typing import ArrayLike

import resfo.types as res_types
from resfo.array_entry import ResArray
from resfo.errors import ResfoParsingError
from resfo.types import ArrayValue


def drop_while_space(stream: IO[str]) -> None:
    """
    Forwards the stream to the first non-space character.
    Assumes the stream is in text mode.
    """
    first_non_space = stream.tell()
    while stream.read(1).isspace():
        first_non_space = stream.tell()
    stream.seek(first_non_space)


class FormattedArray(ResArray):
    """
    An array entry in an formatted res file.
    """

    def read_array(self) -> ArrayValue:
        if self._keyword is None:
            self._read()
        return self._read_array()

    def update(
        self, *, keyword: Optional[str] = None, array: Optional[ArrayLike] = None
    ) -> None:
        raise NotImplementedError("Update is not implemented for formatted files")

    def _read_logical(self) -> bool:
        """
        Read one fortran logical ('T' or 'F') from
        the beginning of the stream and returns it.
        """
        drop_while_space(self.stream)
        current_char = self.stream.read(1)
        if current_char == "T":
            return True
        elif current_char == "F":
            return False
        else:
            raise ResfoParsingError(f"Could not parse logical value {current_char}")

    def _read_quote_separated(self) -> str:
        """
        Read one single-quote delimited string from the
        start of the stream, and returns it.
        """
        drop_while_space(self.stream)
        current_char = self.stream.read(1)
        if current_char != "'":
            raise ResfoParsingError(
                f'Expected "\'" before keyword, got {current_char} at {self.stream.tell()}'
            )
        word = ""

        current_char = self.stream.read(1)
        while current_char != "'":
            word += current_char
            current_char = self.stream.read(1)
            if not current_char:
                raise ResfoParsingError(
                    f"Reached end-of-file while reading keyword {word}"
                )

        return word

    def _read_number(self) -> str:
        """
        Read one number from the stream and returns it as a string.
        """
        drop_while_space(self.stream)
        word = ""

        current_char = self.stream.read(1)
        while current_char in "0123456789DE.+-":
            word += current_char
            current_char = self.stream.read(1)
            if not current_char:
                raise ResfoParsingError(
                    f"Reached end-of-file while reading number {word} at {self.stream.tell()}"
                )

        if not current_char.isspace() and current_char:
            raise ResfoParsingError(
                f"Expected space after number at {self.stream.tell()} got {current_char}"
            )

        if not word:
            raise ResfoParsingError(f"Could not read number at {self.stream.tell()}")

        return word.replace("D", "E")

    def _read_keyword(self) -> None:
        self._keyword = self._read_quote_separated()

    def _read_type(self) -> None:
        self._type = self._read_quote_separated().encode("ascii")

    def _read_length(self) -> None:
        drop_while_space(self.stream)
        self._length = np.fromfile(
            self.stream,
            sep=" ",
            count=1,
            dtype=np.int32,
        )[0]

    def _read_array(self) -> ArrayValue:
        assert self._data_start is not None
        assert self._type is not None
        assert self._length is not None
        self.stream.seek(self._data_start)
        drop_while_space(self.stream)
        if self._type == b"MESS":
            return res_types.MESS
        if res_types.is_character_type(self._type):
            return np.array([self._read_quote_separated() for _ in range(self._length)])
        elif self._type == b"LOGI":
            return np.array([self._read_logical() for _ in range(self._length)])
        else:
            return np.array(
                [self._read_number() for _ in range(self._length)],
                dtype=res_types.to_np_type(self._type),
            )

    def _read(self) -> None:
        """
        See resfo.array_entry.ResArray._read()
        """
        self.stream.seek(self.start)
        drop_while_space(self.stream)
        if not self.stream.read(1):
            self._is_eof = True
            return
        else:
            self.stream.seek(self.start)
        self._read_keyword()
        self._read_length()
        self._read_type()
        self._data_start = self.stream.tell()
        self._read_array()



from .errors import ResfoParsingError, ResfoWriteError
from .format import Format
from .read import lazy_read, read
from .types import MESS, ArrayValue, MessType
from .version import __version__
from .write import write

__author__ = "Equinor"
__email__ = "fg_sib-scout@equinor.com"

__all__ = [
    "read",
    "lazy_read",
    "write",
    "Format",
    "MESS",
    "MessType",
    "ResfoParsingError",
    "ResfoWriteError",
    "ArrayValue",
    "__version__",
]



import numpy as np
from hypothesis import HealthCheck, given, settings
from numpy.testing import assert_allclose

from resfo import MESS, read, write

from .generators import formats, resfo_datas


def same_keyword(a, b):
    return a.strip() == b.strip()


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], print_blob=True)
@given(file_format=formats, data=resfo_datas)
def test_read_write(filelike, file_format, data):
    write(filelike, data, fileformat=file_format)
    for (kw, arr), (okw, oarr) in zip(data, read(filelike), strict=True):
        assert same_keyword(kw, okw)
        if arr is MESS and oarr is MESS:
            continue
        assert (arr is MESS) == (oarr is MESS)
        if len(arr) == 0 and len(oarr) == 0:
            continue
        if isinstance(arr[0], bytes):
            arr = np.array([el.decode("ascii") for el in arr])
        if isinstance(oarr[0], bytes):
            oarr = np.array([el.decode("ascii") for el in oarr])

        if np.issubdtype(arr.dtype, np.inexact):
            assert_allclose(arr, oarr)
        else:
            assert np.array_equal(arr, oarr)



import pytest


@pytest.fixture(params=["file", "path"])
def filelike(request, tmp_path):
    file = tmp_path / "testpath.txt"
    if request.param == "file":
        return str(file)
    if request.param == "path":
        return file



import hypothesis.strategies as st
import pytest
from hypothesis import given

from resfo._unformatted.common import bytes_in_array, group_len, item_size


def test_group_len():
    assert group_len(b"C032") == 105
    assert group_len(b"CHAR") == 105
    assert group_len(b"INTE") == 1000
    assert group_len(b"REAL") == 1000
    assert group_len(b"LOGI") == 1000
    assert group_len(b"DOUB") == 1000
    assert group_len(b"MESS") == 1000
    assert group_len(b"x321") == 1000


def test_item_size():
    assert item_size(b"C032") == 32
    assert item_size(b"CHAR") == 8
    assert item_size(b"INTE") == 4
    assert item_size(b"REAL") == 4
    assert item_size(b"LOGI") == 4
    assert item_size(b"DOUB") == 8
    assert item_size(b"MESS") == 0
    assert item_size(b"x231") is None


def test_bytes_in_array():
    assert bytes_in_array(1000, b"DOUB") == 8008
    assert bytes_in_array(1000, b"REAL") == 4008
    assert bytes_in_array(900, b"REAL") == 900 * 4 + 8
    assert bytes_in_array(1100, b"REAL") == 1000 * 4 + 8 + 100 * 4 + 8
    assert bytes_in_array(0, b"REAL") == 0


@pytest.mark.parametrize(
    "item_type",
    [
        b"C032",
        b"CHAR",
        b"INTE",
        b"REAL",
        b"LOGI",
        b"DOUB",
    ],
)
@given(length=st.integers(min_value=1, max_value=104))
def test_bytes_in_array_one_group(item_type, length):
    assert bytes_in_array(length, item_type) == 8 + length * item_size(item_type)



"""
The types of elements in the res output files are
INTE for 32-bit integers, REAL for 32-bit floats,
LOGI for 32-bit fortran logicals, DOUB for 64-bit floats,
CHAR for 8 byte ascii strings, and C0nn for nn byte ascii strings.

All numerical types are big endian.

Additionally there are two types that signal something other than
what is contained in the arrays. X231 means that the following array
has more than 2**31 elements so the length is split into two headers.

The other is MESS type which signals a 0 length array, so the type
of elements is irrelevant.


A res type is represented by a byte string of ascii characters
having one of the above mentioned keywords.

resfo converts res types to corresponding numpy dtypes
and back. Some numpy dtypes will give a loss of precision and
that results in a warning.

"""

import warnings
from typing import Any, Union

import numpy as np
import numpy.typing as npt
from typing_extensions import TypeAlias

# np dtype for res types with fixed width
static_dtypes: dict[bytes, np.dtype[Any]] = {
    b"INTE": np.dtype(np.int32).newbyteorder(">"),
    b"REAL": np.dtype(np.float32).newbyteorder(">"),
    b"LOGI": np.dtype(np.int32).newbyteorder(">"),
    b"DOUB": np.dtype(np.float64).newbyteorder(">"),
    b"CHAR": np.dtype("|S8"),
}


class MessType:
    """
    The MESS value is a sentinell object used to signal that the type of the
    array that is to be written should be an empty array of type MESS.
    """

    pass


MESS = MessType()


ArrayValue: TypeAlias = Union[npt.NDArray[Any], MessType]


def to_np_type(type_keyword: bytes) -> Union[np.dtype[Any], None]:
    """
    :param type_keyword: A bytestring of a res type.
    :returns: A np dtype corresponding to the given
        res type.
    """
    if type_keyword[0:2] == b"C0":
        return np.dtype("|S" + type_keyword[2:4].decode("ascii"))
    return static_dtypes.get(type_keyword)


def from_np_dtype(array: ArrayValue) -> bytes:
    """
    :param array: A numpy array
    :returns: The corresponding res type for the
        given numpy array's dtype.
    """
    if isinstance(array, MessType):
        return b"MESS"
    dtype = array.dtype
    if dtype == "object":
        return b"MESS"
    if dtype in [np.dtype(np.int32), np.dtype(np.int32).newbyteorder(">")]:
        return b"INTE"
    if dtype in [np.dtype(np.int64), np.dtype(np.int64).newbyteorder(">")]:
        warnings.warn("downcasting numpy int64 to int32 for res file.", stacklevel=1)
        return b"INTE"
    if dtype in [np.dtype(np.float32), np.dtype(np.float32).newbyteorder(">")]:
        return b"REAL"
    if dtype in [np.dtype(np.float64), np.dtype(np.float64).newbyteorder(">")]:
        return b"DOUB"
    if dtype == np.dtype(np.bool_):
        return b"LOGI"
    if dtype == np.dtype("|S8"):
        return b"CHAR"
    if dtype == np.dtype("|S") or dtype == np.dtype("U"):
        max_len = max(len(s) for s in array)
        if max_len > 99:
            raise ValueError("res filetype only supports string length up to 99")
        if max_len == 8:
            return b"CHAR"
        return b"C0" + str(max_len).zfill(2).encode("ascii")
    if dtype.char == "U" and 0 < dtype.itemsize < 4 * 99:
        size = dtype.itemsize // 4
        if size == 8:
            return b"CHAR"
        return b"C0" + str(size).zfill(2).encode("ascii")
    if dtype.char == "S" and 0 < dtype.itemsize < 99:
        size = dtype.itemsize
        if size == 8:
            return b"CHAR"
        return b"C0" + str(size).zfill(2).encode("ascii")
    raise ValueError(f"Could not convert numpy type {dtype} in {array}")


def is_valid_type(type_str: bytes) -> bool:
    """
    :returns: Whether the given byte string is a valid res type.
    """
    if type_str in static_dtypes:
        return True
    if type_str in [b"X231", b"MESS"]:
        return True
    return type_str[0:2] == b"C0" and all(48 <= kw <= 57 for kw in type_str[2:4])


def is_character_type(type_str: bytes) -> bool:
    """
    :returns: Whether the given byte string as ascii is
        a string res type, ie. b"CHAR" or b"C016".
    """
    return bool(type_str) and type_str[0:1] == b"C"



import io
from enum import Enum, auto, unique
from os import PathLike
from typing import IO, Any, Tuple, Union


@unique
class Format(Enum):
    """
    The format of an res file, either FORMATTED for ascii
    or UNFORMATTED for binary.
    """

    FORMATTED = auto()
    UNFORMATTED = auto()


def guess_format(filelike: Union[str, PathLike[str], IO[Any]]) -> Format:
    """
    Guess the format of a given filelike.

    If given a filename or pathlib.Path the guess is based on
    seeing if the first 4 bytes are a fortran marker. If given
    a stream, assumes byte stream means Format.FORMATTED and
    text stream means Format.UNFORMATTED.

    :param filelike: Either a filename, pathlib.Path or opened stream.
    :returns: Either Format.FORMATTED or Format.UNFORMATTED.
    """
    if isinstance(filelike, (PathLike, str)):
        with open(filelike, "rb") as file_handle:
            if int.from_bytes(file_handle.read(4), byteorder="big", signed=True) == 16:
                return Format.UNFORMATTED
            else:
                return Format.FORMATTED
    goback = filelike.tell()
    bytes_or_str = filelike.read(1)
    filelike.seek(goback)
    if isinstance(bytes_or_str, bytes):
        return Format.UNFORMATTED
    else:
        return Format.FORMATTED


def get_stream(
    filepath: Union[str, PathLike[str], IO[Any]], fileformat: Format, mode: str = "r"
) -> Tuple[IO[Any], bool]:
    """
    Openes the given file with the correct mode (text or binary)
    based on fileformat.
    :param filepath: Either a filename or pathlib.Path
    :param fileformat: A resfo.Format.
    :returns: The opened file.
    """
    if isinstance(filepath, (str, PathLike)):
        if fileformat == Format.FORMATTED:
            return open(filepath, mode + "t"), True  # noqa: SIM115
        else:
            return open(filepath, mode + "b"), True  # noqa: SIM115
    else:
        return filepath, False


def check_correct_mode(stream: IO[Any], fileformat: Format) -> None:
    """
    Checks that the stream is the correct mode (text or binary) for the given
    fileformat.
    """
    if isinstance(stream, io.TextIOBase) and fileformat == Format.UNFORMATTED:
        raise ValueError(
            "Formatted file was opened in byte reading mode, should be text mode"
        )



import io

import numpy as np

import resfo._formatted.read as fread


def test_drop_while_space():
    buf = io.StringIO("  a  ")
    fread.drop_while_space(buf)
    assert buf.read(1) == "a"
    fread.drop_while_space(buf)
    assert not buf.read(1)


def test_formatted_array_eof():
    buf = io.StringIO("")
    arr = fread.FormattedArray(buf)
    assert arr.is_eof


def test_simple_array_read(tmp_path):
    contents = "'SPECGRID' 4 'INTE' 40 64 14 1 "

    with (tmp_path / "test.txt").open("w") as f:
        f.write(contents)

    with (tmp_path / "test.txt").open("r") as f:
        arr = fread.FormattedArray(f)

        assert arr.read_type() == b"INTE"
        assert arr.read_length() == 4
        assert arr.read_keyword() == "SPECGRID"
        assert np.array_equal(arr.read_array(), [40, 64, 14, 1])


def test_simple_char_array_read(tmp_path):
    contents = "'CHARARR1' 2 'CHAR' 'HELLO   ' 'WORLD   ' "

    with (tmp_path / "test.txt").open("w") as f:
        f.write(contents)

    with (tmp_path / "test.txt").open("r") as f:
        arr = fread.FormattedArray(f)

        assert arr.read_type() == b"CHAR"
        assert arr.read_length() == 2
        assert arr.read_keyword() == "CHARARR1"
        assert np.array_equal(arr.read_array(), ["HELLO   ", "WORLD   "])


def test_simple_logi_array_read(tmp_path):
    contents = "'KEYWORD1' 2 'LOGI' T F "

    with (tmp_path / "test.txt").open("w") as f:
        f.write(contents)

    with (tmp_path / "test.txt").open("r") as f:
        arr = fread.FormattedArray(f)

        assert arr.read_type() == b"LOGI"
        assert arr.read_length() == 2
        assert arr.read_keyword() == "KEYWORD1"
        assert np.array_equal(arr.read_array(), [True, False])



import numpy as np
import pytest

from resfo._formatted.read import FormattedArray
from resfo._formatted.write import formatted_write


@pytest.mark.filterwarnings("ignore:downcasting")
@pytest.mark.parametrize(
    "data",
    [
        [("KEYWORD1", [1, 2, 3])],
        [("KEYWORD1", [1, 2, 3]), ("22222222", np.array([1.1, 2.1, 3.1]))],
        [
            ("KEYWORD1", np.array([1, 2, 3])),
            ("KEYWORD2", [1.0, 2.0, 3.0]),
        ],
        [
            ("KEYWORDS", np.array(["HELLO WORLD"], dtype="|S")),
            ("KEYWORDU", np.array(["HELLO WORLD"], dtype="|U")),
            ("KEYWORDU", np.array(["HELLO WORLD"], dtype="|U11")),
        ],
    ],
)
def test_formatted_write(data, tmp_path):
    with (tmp_path / "test.data").open("wt") as fh:
        formatted_write(fh, data)

    with (tmp_path / "test.data").open("rt") as fh:
        out_records = list(FormattedArray.parse(fh))
        out_keywords = [r.read_keyword() for r in out_records]
        out_arrays = [r.read_array() for r in out_records]
        out_lengths = [r.read_length() for r in out_records]

    assert out_lengths == [len(arr) for arr in out_arrays]

    for (kw, arr), okw, oarr in zip(data, out_keywords, out_arrays, strict=True):
        assert kw == okw
        if isinstance(arr[0], bytes):
            assert np.array_equal([el.decode("ascii") for el in arr], oarr)
        else:
            assert np.array_equal(arr, oarr)



from os import PathLike
from typing import IO, Any, Iterable, Tuple, Union

import numpy.typing as npt

from resfo._formatted.write import formatted_write
from resfo._unformatted.write import unformatted_write
from resfo.format import Format, check_correct_mode, get_stream

from .types import MessType


def write(
    filelike: Union[str, PathLike[str], IO[Any]],
    contents: Iterable[Tuple[str, Union[npt.ArrayLike, MessType]]],
    fileformat: Format = Format.UNFORMATTED,
) -> None:
    """
    Write the given contents to the given file in res format.

    :param filelike: Either filename, pathlib.Path or stream
        to write file to. For fileformat=Format.UNFORMATTED the
        stream must be in binary mode and for fileformat=Format.FORMATTED
        in text mode.
    :param contents: list or iterable of tuples (kw, arr) where keyword
        is the keyword, and arr is a numpy arraylike of values. The
        keyword must have exactly 8 characters and the type of the array
        will be converted according to resfo.types.to_np_type
    :param fileformat: Either resfo.Format.FORMATTED for ascii
        format or resfo.Format.UNFORMATTED for binary format.

    :raises resfo.ResfoWriteError: If the given contents cannot be
        written to an res file.

    .. note::
        If given a file to be open (as opposed to a stream), the errors
        (various `IOError` s) associated with the default behavior of the
        built-in `open()` function may be raised.

        When given a stream, the exceptions associated with the stream will
        pass through.
    """
    stream, didopen = get_stream(filelike, fileformat, mode="w")

    check_correct_mode(stream, fileformat)

    if fileformat == Format.FORMATTED:
        formatted_write(stream, contents)
    else:
        unformatted_write(stream, contents)

    if didopen:
        stream.close()



import io

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings

from resfo import MESS, lazy_read, read, write

from .generators import float_arrays, keywords, resfo_datas


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(data=st.lists(st.tuples(keywords, float_arrays)))
def test_update(filelike, data):
    data_zeros = [(kw, np.zeros(array.shape, dtype=array.dtype)) for kw, array in data]

    write(filelike, data_zeros)
    with open(filelike, "br+") as stream:
        for entry, (kw, arr) in zip(lazy_read(stream), data, strict=True):
            assert kw == entry.read_keyword()
            entry.update(keyword="NEW_NAME")
            assert entry.read_array().tolist() == [0] * arr.size
            assert entry.read_keyword() == "NEW_NAME"
            entry.update(array=arr)
            assert entry.read_array().tolist() == pytest.approx(arr.tolist())


def test_update_mismatching_array_raises(filelike):
    write(filelike, [("KEYWORD1", [True, False])])

    with open(filelike, "br+") as stream:
        reader = lazy_read(stream)
        entry = next(reader)

        with pytest.raises(ValueError, match="different size"):
            entry.update(array=[True])

        with pytest.raises(ValueError, match="different type"):
            entry.update(array=["a", "b"])


def test_update_incorrect_mode_raises(filelike):
    write(filelike, [("KEYWORD1", [True, False])])

    with open(filelike, "br") as stream:
        entry = next(lazy_read(stream))
        with pytest.raises(io.UnsupportedOperation):
            entry.update(array=[False, False])


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(data=resfo_datas)
def test_update_preserves_type(filelike, data):
    write(filelike, data)
    with open(filelike, "br+") as stream:
        for (kw, arr), entry in zip(data, lazy_read(stream), strict=True):
            entry.update(keyword=kw, array=arr)

    for (kw, arr), (okw, oarr) in zip(data, read(filelike), strict=True):
        assert kw == okw
        if arr is MESS:
            continue
        if arr.dtype.type == np.str_:
            # Allow reading to result in bytestring instead of string
            assert oarr.dtype.type in [np.dtype("S8").type, np.dtype("<U8").type]
        else:
            assert arr.dtype.type == oarr.dtype.type



from abc import ABC, abstractmethod
from typing import IO, Any, Iterator, Optional, Union

from numpy.typing import ArrayLike
from typing_extensions import Self

from .types import ArrayValue


class ResArray(ABC):
    """
    An array entry in a res file.

    This class is not ment to be constructed directly, but rather
    generated e.g. :py:meth:`resfo.lazy_read`.
    """

    __slots__ = (
        "start",
        "stream",
        "_keyword",
        "_length",
        "_type",
        "_data_start",
        "_is_eof",
        "_array",
    )

    def __init__(self, stream: IO[Any]) -> None:
        """
        :param stream: The opened res file with stream.peek() at
            the start of the res array.
        """
        self.start = stream.tell()
        self.stream = stream

        self._keyword: Optional[str] = None
        self._length: Optional[int] = None
        self._type: Optional[bytes] = None
        self._data_start: Optional[int] = None
        self._is_eof = False
        self._array: Optional[ArrayValue] = None

    @property
    def is_eof(self) -> bool:
        """
        Whether the position of the array is actually
        at the end of the file, in which case its keyword
        and array is None.
        """
        if self._keyword is None:
            self._read()
        return self._is_eof

    def read_keyword(self) -> str:
        """
        Read the keyword from the res file.

        :returns: The keyword as a 8 character string.
        """
        if self._keyword is None:
            self._read()
        return self._keyword  # type: ignore

    def read_length(self) -> int:
        """
        Read the length from the res file.

        :returns: The length of the array in number of entries.
        """
        if self._length is None:
            self._read()
        return self._length  # type: ignore

    @abstractmethod
    def read_array(self) -> ArrayValue:
        """
        Read the array from the unformatted res file.

        :returns: numpy array of values.
        """
        pass

    def read_type(self) -> Union[str, bytes]:
        """
        The type given in the header of the array.
        In case of unformatted file this will be bytes, for
        unformatted files it is str.
        """
        if self._type is None:
            self._read()
        return self._type  # type: ignore

    @abstractmethod
    def update(
        self, *, keyword: Optional[str] = None, array: Optional[ArrayLike] = None
    ) -> None:
        """
        Updates the entry with the given new data.

        :param new_keyword: The new keyword, must be 8 characters, defaults to no change.
        :param new_array: The new array, must be of same type and size
            as existing array, defaults to no change.

        :raises ValueError: If array does not have the same type or length as
            existing array.
        :raises io.UnsupportedOperation: If the underlying stream has been opened
            with the wrong mode (file must be opened with "r+" in order to correctly update)
        """
        pass

    @abstractmethod
    def _read(self) -> None:
        """
        Reads the array entry. Guarantees that after, either entry.is_eof() and
        the stream.peek() is at the end of the file or stream.peek() will be at
        the start of the next record.
        """
        pass

    @classmethod
    def parse(cls, stream: IO[Any]) -> Iterator[Self]:
        """
        Parse an res file from the given opened file handle.

        Is a generator of ResArrays
        """
        record = cls(stream)

        record._read()

        while not record._is_eof:
            new_record = cls(stream)
            yield record
            new_record._read()
            record = new_record



from io import BytesIO, StringIO

import pytest

from resfo import Format, ResfoParsingError, ResfoWriteError, read, write


def test_invalid_formatted_content():
    file_contents = StringIO("Not valid res content")
    with pytest.raises(ResfoParsingError, match='Expected "\'"'):
        read(file_contents)


def test_invalid_unformatted_content():
    file_contents = BytesIO(b"\x00" * 100)
    with pytest.raises(ResfoParsingError, match="Unexpected size of record"):
        read(file_contents)


@pytest.mark.parametrize(
    "buffer, format", [(BytesIO(), Format.UNFORMATTED), (StringIO(), Format.FORMATTED)]
)
def test_invalid_read_content(buffer, format):
    with pytest.raises(ResfoWriteError, match="Could not convert"):
        write(buffer, [("FILEHEAD", "a" * 100)], format)


def test_invalid_keyword():
    with pytest.raises(ResfoWriteError, match="keywords"):
        write(StringIO(), [("'", ["a" * 8])], Format.FORMATTED)
